import click
from pathlib import Path
from stubbles.inputs import read_file_to_dict, language
from stubbles.templates import populate


@click.group()
@click.option('--template', type=click.Path(exists=True, dir_okay=False), required=True)
@click.option('--replacements', type=click.Path(exists=True, dir_okay=False), required=True)
@click.option('--output-dir', type=click.Path(dir_okay=True))
def cli(template, replacements, output_dir):
    if not output_dir:
        output_dir = Path('generated')

    main(
        template=template,
        replacements_file=replacements,
        output_dir=output_dir
    )


def main(template: Path, replacements_file: Path, output_dir: Path):
    replacements = read_file_to_dict(replacements_file)
    with open(template, 'r') as template_file:
        contents = template_file.read()
        lang = language(str(template))
        new_template = populate(template=contents, replacements=replacements, lang=lang)

    with open(output_dir / 'file.random', 'w') as file:
        file.write(new_template)


if __name__ == '__main__':
    cli()



import pytest
from stubbles.types import Language, extensions


def test_languages_unique():
    unique_extensions = set()
    for extension_list in extensions.values():
        for ext in extension_list:
            unique_extensions.add(ext)

    total_extensions = sum([len(ext) for ext in extensions.values()])
    assert total_extensions == len(unique_extensions)



import pytest
import json
from stubbles.types import Language
from stubbles.inputs import language, read_file_to_dict


@pytest.mark.parametrize(
    'extension,desired', (
        ('.c', Language.C),
        ('.cpp', Language.CPP),
        ('.py', Language.PYTHON),
        ('.java', Language.JAVA),
        ('.tsx', Language.TYPESCRIPT),
        ('.go', Language.GOLANG),

    )
)
def test_extensions(extension, desired):
    actual = language(extension)
    assert actual == desired


def test_read_json_file(tmp_path):
    p = tmp_path / "test.json"
    p.write_text(json.dumps({"key": "value"}))
    assert read_file_to_dict(str(p)) == {"key": "value"}


def test_read_xml_file(tmp_path):
    p = tmp_path / "test.xml"
    p.write_text("<root><key>value</key></root>")
    assert read_file_to_dict(str(p)) == {"key": "value"}


def test_read_yaml_file(tmp_path):
    p = tmp_path / "test.yaml"
    p.write_text("key: value")
    assert read_file_to_dict(str(p)) == {"key": "value"}


def test_read_toml_file(tmp_path):
    p = tmp_path / "test.toml"
    p.write_text("key = \"value\"")
    assert read_file_to_dict(str(p)) == {"key": "value"}


def test_read_ini_file(tmp_path):
    p = tmp_path / "test.ini"
    p.write_text("[section]\nkey=value")
    assert read_file_to_dict(str(p)) == {"section": {"key": "value"}}


def test_read_csv_file(tmp_path):
    p = tmp_path / "test.csv"
    p.write_text("key,value\nkey1,val1")
    assert read_file_to_dict(str(p)) == {"key1": "val1"}


def test_read_unsupported_file(tmp_path):
    p = tmp_path / "test.txt"
    p.write_text("this is a text file")
    with pytest.raises(ValueError):
        read_file_to_dict(str(p))


def test_read_nonexistent_file(tmp_path):
    p = tmp_path / "does_not_exist.xyz"
    with pytest.raises(Exception):
        read_file_to_dict(str(p))



from enum import Enum


class Language(Enum):
    PYTHON = "Python"
    JAVA = "Java"
    C = "C"
    CPP = "C++"
    JAVASCRIPT = "JavaScript"
    CSHARP = "C#"
    RUBY = "Ruby"
    SWIFT = "Swift"
    GOLANG = "Go"
    PHP = "PHP"
    TYPESCRIPT = "TypeScript"
    KOTLIN = "Kotlin"
    RUST = "Rust"
    HTML = "HTML"
    CSS = "CSS"


comments = {
    Language.PYTHON: "#",
    Language.JAVA: "//",
    Language.C: "//",
    Language.CPP: "//",
    Language.JAVASCRIPT: "//",
    Language.CSHARP: "//",
    Language.RUBY: "#",
    Language.SWIFT: "//",
    Language.GOLANG: "//",
    Language.PHP: "//",
    Language.TYPESCRIPT: "//",
    Language.KOTLIN: "//",
    Language.RUST: "//",
    Language.CSS: "//"

}

extensions = {
    Language.PYTHON: ['.py'],
    Language.JAVA: ['.java'],
    Language.C: ['.c', '.h'],
    Language.CPP: ['.cpp', '.cc', '.cxx', '.hpp'],
    Language.JAVASCRIPT: ['.js', '.jsx'],
    Language.CSHARP: ['.cs'],
    Language.RUBY: ['.rb'],
    Language.SWIFT: ['.swift'],
    Language.GOLANG: ['.go'],
    Language.PHP: ['.php'],
    Language.TYPESCRIPT: ['.ts', '.tsx'],
    Language.KOTLIN: ['.kt', '.kts'],
    Language.RUST: ['.rs'],
    Language.CSS: ['.css', '.scss', '.sass'],
}



from stubbles.types import Language, extensions
import json
import os
import csv
import configparser
import xml.etree.ElementTree as ET
import yaml
import toml


def language(extension: str) -> Language:
    for lang, exts in extensions.items():
        if extension in exts:
            return lang

    raise ValueError(f"Language not found for extension {extension}.")


def read_file_to_dict(filepath):
    file_ext = os.path.splitext(filepath)[-1]
    output_dict = {}

    if file_ext == '.json':
        with open(filepath, 'r') as f:
            output_dict = json.load(f)

    elif file_ext == '.xml':
        tree = ET.parse(filepath)
        root = tree.getroot()
        for elem in root:
            output_dict[elem.tag] = elem.text

    elif file_ext == '.yaml' or file_ext == '.yml':
        with open(filepath, 'r') as f:
            output_dict = yaml.safe_load(f)

    elif file_ext == '.toml':
        output_dict = toml.load(filepath)

    elif file_ext == '.ini':
        config = configparser.ConfigParser()
        config.read(filepath)
        for section in config.sections():
            output_dict[section] = {}
            for key, val in config.items(section):
                output_dict[section][key] = val

    elif file_ext == '.csv':
        with open(filepath, 'r') as f:
            reader = csv.reader(f)
            header = next(reader, None)
            if header:
                for row in reader:
                    key = row[0]
                    value = row[1]
                    output_dict[key] = value

    else:
        raise ValueError(f"Unsupported file extension: {file_ext}")

    return output_dict



import pytest
from stubbles.templates import _lines, _strip, _stubs, _whitespace, validate, populate, _is_stubble_line
from stubbles.types import Language


@pytest.mark.parametrize(
    'line,lang,desired',
    (
        ('// {{hello}}', Language.CPP, True),
        ('# Yolo', Language.PYTHON, False),
        ('# {{Yolo}}', Language.PYTHON, True),
        ('# Serial.printlln({{key}})', Language.PYTHON, True),
    )
)
def test_is_stubble_line(line, lang, desired):
    actual = _is_stubble_line(line=line, lang=lang)
    assert actual == desired


def test_lines_python():
    template = """
    # {{replacement_code}}
    print("Hello, world!")  # This is a comment
    #
# {{other_replacement_code}}
    \t#
        # This is an indented comment
        print("Indented code")
    \t\t\t#
"""

    actual = _lines(template=template, lang=Language.PYTHON)
    desired = [
        "    # {{replacement_code}}",
        "# {{other_replacement_code}}"
    ]
    assert actual == desired


@pytest.mark.parametrize(
    'string,desired,lang', (
        ('  # {{hello_there}}  ', '{{hello_there}}', Language.PYTHON),
        ('# {{hihi}}   ', '{{hihi}}', Language.PYTHON),
        ('     // {{hihi}}   ', '{{hihi}}', Language.CPP),
    )
)
def test_strip(string, desired, lang):
    actual = _strip(line=string, lang=lang)
    assert actual == desired


@pytest.mark.parametrize(
    'string,desired,lang', (
            ('  # {{hello_there}}  ', '  ', Language.PYTHON),
            ('# {{hihi}}   ', '', Language.PYTHON),
            ('     // {{hihi}}   ', '     ', Language.CPP),
    )
)
def test_whitespace(string, desired, lang):
    actual = _whitespace(line=string, lang=lang)
    assert actual == desired


@pytest.mark.parametrize(
    'string,desired,lang', (
            ('  # {{hello_there}}  ', ['hello_there'], Language.PYTHON),
            ('# {{hihi}}   ', ['hihi'], Language.PYTHON),
            ('     // {{hihi}}   ', ['hihi'], Language.CPP),
            ('     // {{hihi}}  {{yesyes}}  ', ['hihi', 'yesyes'], Language.CPP),
            ('     // hihi  yesyes  ', [], Language.CPP),
    )
)
def test_stubs_single(string, desired, lang):
    actual = _stubs(line=string, lang=lang)
    assert actual == desired


def test_stubs_multi():
    string = '# {{hi}} {{there}}'
    actual = _stubs(line=string, lang=Language.PYTHON)
    desired = ['hi', 'there']
    assert actual == desired


def test_validate_raises():
    template = '{{hello}}'
    replacements = {'goodbye': 'hello'}
    with pytest.raises(AssertionError):
        validate(template=template, lang=Language.PYTHON, replacements=replacements)


def test_validate_ok():
    template = '{{hello}}'
    replacements = {'hello': 'hello'}
    validate(template=template, lang=Language.PYTHON, replacements=replacements)


def test_replace():
    input_template = """
Hello there partner    
# {{key1}}
    # {{key2}}
\t#{{key3}}
"""
    desired = """
Hello there partner    
replacement1
    replacement2
\treplacement3
"""
    replacements = {
        'key1': 'replacement1',
        'key2': 'replacement2',
        'key3': 'replacement3',
    }
    actual = populate(template=input_template, lang=Language.PYTHON, replacements=replacements)
    assert actual == desired


def test_replace_inline():
    template = '// Serial.println({{key}});'
    replacements = {'key': 'val'}
    lang = Language.CPP
    actual = populate(template=template, lang=lang, replacements=replacements)
    desired = 'Serial.println(val);'
    assert actual == desired




from stubbles.types import Language, comments
from typing import List, Dict, Mapping, Tuple, Any
from copy import copy
import itertools


def populate(template: str, lang: Language, replacements: Dict[str, str]) -> str:
    all_lines = template.split('\n')
    new_lines = []
    for line in all_lines:
        if _is_stubble_line(line=line, lang=lang):
            replaced = _replace(line=line, lang=lang, replacements=replacements)
            stripped = _strip(line=replaced, lang=lang)
            new_line = _whitespace(line=line, lang=lang) + stripped
        else:
            new_line = line
        new_lines.append(new_line)

    populated_template = '\n'.join(new_lines)

    return populated_template


def validate(template: str, lang: Language, replacements: Mapping[str, str]):
    missing_keys = []

    all_lines = _lines(template=template, lang=lang)
    all_stubbles = _flatten([_stubs(line=line, lang=lang) for line in all_lines])

    for stub in all_stubbles:
        if stub not in replacements.keys():
            missing_keys.append(stub)
    if missing_keys:
        raise AssertionError(f"Replacements missing keys: {missing_keys}")

    for key, val in replacements.items():
        if not isinstance(val, str):
            raise AssertionError("Replacements must contain only strings as values")



START_STUBBLE = "{{"
END_STUBBLE = "}}"


def _flatten(nested_list: List[List[Any]]):
    return list(itertools.chain.from_iterable(nested_list))


def _stub_indices(string: str, inclusive=True) -> Tuple[int, int]:
    if START_STUBBLE not in string or END_STUBBLE not in string:
        return 0, 0

    start_index = string.find(START_STUBBLE)
    end_index = string.find(END_STUBBLE)

    if inclusive:
        end_index += len(END_STUBBLE)
    else:
        start_index += len(START_STUBBLE)

    return start_index, end_index


def _strip(line: str, lang: Language) -> str:
    comment = comments[lang]
    stripped_line = line.lstrip()  # Remove leading whitespaces
    bare_line = stripped_line.lstrip(comment).strip()
    return bare_line


def _is_stubble_line(line: str, lang: Language) -> bool:
    stripped_line = _strip(line=line, lang=lang)

    if START_STUBBLE in stripped_line and END_STUBBLE in stripped_line:
        return True

    return False


def _lines(template: str, lang: Language) -> List[str]:
    all_lines = template.split('\n')
    stubble_lines = []

    for line in all_lines:
        if _is_stubble_line(line=line, lang=lang):
            stubble_lines.append(line)

    return stubble_lines


def _stubs(line: str, lang: Language) -> List[str]:
    return _stubble_internal(string=line, lang=lang, stubs=[])


def _stubble_internal(string: str, lang: Language, stubs: List[str]):
    new_list = copy(stubs)
    start, end = _stub_indices(string=string, inclusive=True)
    stub = string[start:end]
    if stub:
        no_brackets_stub = stub.removeprefix(START_STUBBLE).removesuffix(END_STUBBLE)
        no_whitespace_stub = no_brackets_stub.strip()
        new_list.append(no_whitespace_stub)
        new_string = string[end:]
        new_list = _stubble_internal(string=new_string, lang=lang, stubs=new_list)

    return new_list


def _whitespace(line: str, lang: Language) -> str:
    if not _is_stubble_line(line=line, lang=lang):
        raise ValueError("Line must be a valid stubble line.")

    comment = comments[lang]
    spaces, *rest = line.split(comment)
    return spaces



def _replace(line: str, lang: Language, replacements: Mapping[str, str]) -> str:
    stubs_in_line = _stubs(line=line, lang=lang)
    remaining_line = line
    for stub in stubs_in_line:
        start, end = _stub_indices(string=line, inclusive=True)
        remaining_line = remaining_line[:start] + replacements[stub] + remaining_line[end:]

    return remaining_line


